""" SACAD cover sources. """

from sacad.sources.amazoncd import AmazonCdCoverSource, AmazonCdCoverSourceResult  # noqa: F401
from sacad.sources.amazondigital import AmazonDigitalCoverSource, AmazonDigitalCoverSourceResult  # noqa: F401
from sacad.sources.deezer import DeezerCoverSource, DeezerCoverSourceResult  # noqa: F401
from sacad.sources.google_images import GoogleImagesCoverSourceResult, GoogleImagesWebScrapeCoverSource  # noqa: F401
from sacad.sources.lastfm import LastFmCoverSource, LastFmCoverSourceResult  # noqa: F401
